/* redirected_key() conversion */

#include <tt.h>

#include <id.h>
static char s1[] = "Begin escape character ";
static unsigned begin_escape_character = '<';
static char s2[] = "End escape character ";
static unsigned end_escape_character = '>';
static char sx[] = "";

static char HELP[] = "\r\n"
  "REDKEY  sourcefile  destinationfile [*]\r\n"
  "\r\n"
  "Copy the source file to the destination file, removing control\r\n"
  "characters and converting certain escape sequences to\r\n"
  "sequences acceptable to redirected_key().  The optional third\r\n"
  "parameter means display the source file as it is processed.\r\n"
  "\r\n"
  "Type REDKEY * to see the escape sequences.\r\n"
  "\r\n"
  "CUSTOM can be used to change the escape sequence delimiters.\r\n";

#define NUMBER_OF_ESCAPES (sizeof(escape)/sizeof(struct escape_struct))

static struct escape_struct
{
  char *name;
  unsigned code;
} escape[] =
{
  " ",               0,
  "Back",            Backspace,
  "Del",             Del,
  "Down",            Down,
  "End",             End,
  "Enter",           Enter,
  "Esc",             Esc,
  "Home",            Home,
  "Ins",             Ins,
  "Left",            Left,
  "PgDn",            PgDn,
  "PgUp",            PgUp,
  "Right",           Right,
  "Space",           Space,
  "Tab",             Tab,
  "Up",              Up,
  "F1",              F1,
  "F2",              F2,
  "F3",              F3,
  "F4",              F4,
  "F5",              F5,
  "F6",              F6,
  "F7",              F7,
  "F7",              F7,
  "F8",              F8,
  "F9",              F9,
  "F10",             F10,
  "Ctrl A",          CtrlA,
  "Ctrl B",          CtrlB,
  "Ctrl C",          CtrlC,
  "Ctrl D",          CtrlD,
  "Ctrl E",          CtrlE,
  "Ctrl F",          CtrlF,
  "Ctrl G",          CtrlG,
  "Ctrl H",          CtrlH,
  "Ctrl I",          CtrlI,
  "Ctrl J",          CtrlJ,
  "Ctrl K",          CtrlK,
  "Ctrl L",          CtrlL,
  "Ctrl M",          CtrlM,
  "Ctrl N",          CtrlN,
  "Ctrl O",          CtrlO,
  "Ctrl P",          CtrlP,
  "Ctrl Q",          CtrlQ,
  "Ctrl R",          CtrlR,
  "Ctrl S",          CtrlS,
  "Ctrl T",          CtrlT,
  "Ctrl U",          CtrlU,
  "Ctrl V",          CtrlV,
  "Ctrl W",          CtrlW,
  "Ctrl X",          CtrlX,
  "Ctrl Y",          CtrlY,
  "Ctrl Z",          CtrlZ,
  "Ctrl @",          CtrlAt,
  "Ctrl \\",         CtrlBackslash,
  "Ctrl ^",          CtrlCaret,
  "Ctrl [",          CtrlLeftBracket,
  "Ctrl ]",          CtrlRightBracket,
  "Ctrl _",          CtrlUnderscore,
  "Ctrl Back",       CtrlBackspace,
  "Ctrl End",        CtrlEnd,
  "Ctrl Enter",      CtrlEnter,
  "Ctrl Home",       CtrlHome,
  "Ctrl Left",       CtrlLeft,
  "Ctrl PgDn",       CtrlPgDn,
  "Ctrl PgUp",       CtrlPgUp,
  "Ctrl Right",      CtrlRight,
  "Ctrl F1",         CtrlF1,
  "Ctrl F2",         CtrlF2,
  "Ctrl F3",         CtrlF3,
  "Ctrl F4",         CtrlF4,
  "Ctrl F5",         CtrlF5,
  "Ctrl F6",         CtrlF6,
  "Ctrl F7",         CtrlF7,
  "Ctrl F8",         CtrlF8,
  "Ctrl F9",         CtrlF9,
  "Ctrl F10",        CtrlF10,
  "Shift Tab",       ShiftTab,
  "Shift F1",        ShiftF1,
  "Shift F2",        ShiftF2,
  "Shift F3",        ShiftF3,
  "Shift F4",        ShiftF4,
  "Shift F5",        ShiftF5,
  "Shift F6",        ShiftF6,
  "Shift F7",        ShiftF7,
  "Shift F8",        ShiftF8,
  "Shift F9",        ShiftF9,
  "Shift F10",       ShiftF10,
  "Alt 0",           Alt0,
  "Alt 1",           Alt1,
  "Alt 2",           Alt2,
  "Alt 3",           Alt3,
  "Alt 4",           Alt4,
  "Alt 5",           Alt5,
  "Alt 6",           Alt6,
  "Alt 7",           Alt7,
  "Alt 8",           Alt8,
  "Alt 9",           Alt9,
  "Alt A",           AltA,
  "Alt B",           AltB,
  "Alt C",           AltC,
  "Alt D",           AltD,
  "Alt E",           AltE,
  "Alt F",           AltF,
  "Alt G",           AltG,
  "Alt H",           AltH,
  "Alt I",           AltI,
  "Alt J",           AltJ,
  "Alt K",           AltK,
  "Alt L",           AltL,
  "Alt M",           AltM,
  "Alt N",           AltN,
  "Alt O",           AltO,
  "Alt P",           AltP,
  "Alt Q",           AltQ,
  "Alt R",           AltR,
  "Alt S",           AltS,
  "Alt T",           AltT,
  "Alt U",           AltU,
  "Alt V",           AltV,
  "Alt W",           AltW,
  "Alt X",           AltX,
  "Alt Y",           AltY,
  "Alt Z",           AltZ,
  "Alt -",           AltMinus,
  "Alt =",           AltEqual,
  "Alt F1",          AltF1,
  "Alt F2",          AltF2,
  "Alt F3",          AltF3,
  "Alt F4",          AltF4,
  "Alt F5",          AltF5,
  "Alt F6",          AltF6,
  "Alt F7",          AltF7,
  "Alt F8",          AltF8,
  "Alt F9",          AltF9,
  "Alt F10",         AltF10,
};

static jump_buffer file_read_error, file_write_error;
static FILE *source, *destination;
static boolean show_source_file = FALSE;

char program_name[] = "REDKEY (Redirected Key Converter)";

static void fatal_error(char *s)
{
  close_putchar();
  error_exit(s);
}

static int read_character(void)
{
  int c = getc(source);
  if (show_source_file) putchar(c);
  return c;
}

void main(void)
{
  int c;
  escape[0].code = begin_escape_character;
  escape[0].name[0] = begin_escape_character;
  if (command_line_argument(1) == NULL || *command_line_argument(1) == '?')
  {
    putchar_string(HELP);
    close_putchar();
    exit(1);
  }
  if (*command_line_argument(1) == '*')
  {
    unsigned i;
    for (i=0; i<23; i++)
    {
      unsigned j;
      for (j=0; j<6; j++)
      {
        unsigned k = 23*j+i;
        if (k<NUMBER_OF_ESCAPES)
        {
          char *t;
          unsigned n = 11;
          putchar(begin_escape_character);
          for (t=escape[k].name; *t!=0; t++) {putchar(*t); n--;}
          putchar(end_escape_character);
          while (n!=0) {putchar(' '); n--;}
        }
      }
      putchar_string("\r\n");
    }
    close_putchar();
    exit(1);
  }
  if (command_line_argument(2)==NULL)
    fatal_error("Missing destination file");
  if (jumped(file_read_error))
    fatal_error("File read error");
  if (jumped(file_write_error))
    fatal_error("File write error");
  if ((source=fopen(command_line_argument(1), FALSE, 1024, NULL, NULL))==NULL)
  {
    fatal_error("Can't open source file");
  }
  if ((destination=fopen(command_line_argument(2), TRUE, 1024, NULL, NULL)) == 
    NULL)
  {
    fatal_error("Can't open destination file");
  }
  {
    char *p = command_line_argument(3);
    if (p!=NULL && *p == '*') show_source_file = TRUE;
  }
  while ((c=read_character()) != EOF)
  {
    if (c == begin_escape_character)
    {
      static char name[15];
      unsigned i = 0;
      while ((c=read_character()) != end_escape_character)
      {
        if (c==EOF) fatal_error("Escape sequence not terminated");
        if (i < sizeof(name)-1)
          name[i++] = 'a'<=c && c<='z' ? c-('a'-'A') : c;
      }
      name[i] = 0;
      for (i=0; ; i++)
      {
        char *s;
        register char *t;
        if (i==NUMBER_OF_ESCAPES) fatal_error("Unrecognized escape");
        for (s = name, t = escape[i].name; *s != 0; s++, t++)
          if (*s != ('a'<=*t && *t<='z' ? *t-('a'-'A') : *t)) break;
        if (*s == 0) break;        
      }
      {
        unsigned k = escape[i].code;
        if (k>256) putc(0, destination);
        putc(k, destination);
      }
    }
    else if (' '<=c && c<='~')
      putc(c, destination);
  }
  fclose(destination);
  close_putchar();
}
