/* Plain Vanilla TT Library 1.2 header */

void          cdecl __cli__(void);
void          cdecl __sti__(void);
unsigned char cdecl __inportb__(int port);
void          cdecl __outportb__(int port, unsigned char value);
void          cdecl __int__(int interruptnum);

#define NULL     ((void *) 0)
#define FAR_NULL ((void far *) 0)
#define EOF      (-1)
#define TRUE     1
#define FALSE    0

typedef int boolean;

struct date_and_time
{
  char second;
  char minute;
  char hour;
  char day;
  char month;
  int year;
  char day_of_week;
};

struct time
{
  unsigned double_second : 5;
  unsigned minute : 6;
  unsigned hour : 5;
};

struct date
{
  unsigned day : 5;
  unsigned month : 4;
  unsigned year : 7;
};

struct file_information
{
  char reserved_1[21];
  char attribute;
    #define FA_READ_ONLY 0x01
    #define FA_HIDDEN    0x02
    #define FA_SYSTEM    0x04
    #define FA_LABEL     0x08
    #define FA_DIRECTORY 0x10
    #define FA_ARCHIVE   0x20
  struct time time;
  struct date date;
  long size;
  char name[13];
  char reserved_2[85];
};

typedef unsigned jump_buffer[5];

typedef struct
{
  int handle;
  unsigned i, n;
  unsigned flags;
  unsigned buffer_size;
  unsigned *error;
  unsigned *end_of_file;
  unsigned char *buffer;
} FILE;

typedef enum
{
  DRIVE_TYPE_UNKNOWN,
  DRIVE_TYPE_5IN_DD,
  DRIVE_TYPE_5IN_HD,
  DRIVE_TYPE_3IN_DD,
  DRIVE_TYPE_3IN_HD,
} drive_type;

/* device information bit patterns */

  #define IS_DEVICE             0x0080

  /* for device: */

    #define CONSOLE_INPUT_DEVICE  0x0001
    #define CONSOLE_OUTPUT_DEVICE 0x0002
    #define NULL_DEVICE           0x0004
    #define CLOCK_DEVICE          0x0008
    #define RAW_MODE              0x0020
    #define CONTROL_IO            0x4000

  /* for file: */

    #define DRIVE_NUMBER_MASK     0x003F
    #define HAS_BEEN_WRITTEN      0x0040

typedef enum
{
  FS_FILE_OR_DIRECTORY,
  FS_AMBIGUOUS_FILE,
  FS_DIRECTORY,
  FS_DRIVE,
  FS_NOTHING,
  FS_ERROR
} filespecs_type;

struct serial
{
  unsigned base;    
  unsigned char *buffer;
  unsigned index;
  unsigned count;
  unsigned size;
  boolean overflow;
  unsigned irq;
  unsigned mask;
  void far *save_interrupt_vector;
  struct
  {
    unsigned char push_si;
    unsigned char move_si;
    struct serial *pointer;
    unsigned char jmp;
    unsigned handler;
  } interrupt_entry;
};

#define NO_PARITY   0
#define ODD_PARITY  1
#define EVEN_PARITY 3

/* key codes */

#define Alt0              385
#define Alt1              376
#define Alt2              377
#define Alt3              378
#define Alt4              379
#define Alt5              380
#define Alt6              381
#define Alt7              382
#define Alt8              383
#define Alt9              384
#define AltA              286
#define AltB              304
#define AltC              302
#define AltD              288
#define AltE              274
#define AltF              289
#define AltG              290
#define AltH              291
#define AltI              279
#define AltJ              292
#define AltK              293
#define AltL              294
#define AltM              306
#define AltN              305
#define AltO              280
#define AltP              281
#define AltQ              272
#define AltR              275
#define AltS              287
#define AltT              276
#define AltU              278
#define AltV              303
#define AltW              273
#define AltX              301
#define AltY              277
#define AltZ              300
#define AltF1             360
#define AltF2             361
#define AltF3             362
#define AltF4             363
#define AltF5             364
#define AltF6             365
#define AltF7             366
#define AltF8             367
#define AltF9             368
#define AltF10            369
#define AltMinus          386
#define AltEqual          387
#define CtrlA             ('A'-64)
#define CtrlB             ('B'-64)
#define CtrlC             ('C'-64)
#define CtrlD             ('D'-64)
#define CtrlE             ('E'-64)
#define CtrlF             ('F'-64)
#define CtrlG             ('G'-64)
#define CtrlH             ('H'-64)
#define CtrlI             ('I'-64)
#define CtrlJ             ('J'-64)
#define CtrlK             ('K'-64)
#define CtrlL             ('L'-64)
#define CtrlM             ('M'-64)
#define CtrlN             ('N'-64)
#define CtrlO             ('O'-64)
#define CtrlP             ('P'-64)
#define CtrlQ             ('Q'-64)
#define CtrlR             ('R'-64)
#define CtrlS             ('S'-64)
#define CtrlT             ('T'-64)
#define CtrlU             ('U'-64)
#define CtrlV             ('V'-64)
#define CtrlW             ('W'-64)
#define CtrlX             ('X'-64)
#define CtrlY             ('Y'-64)
#define CtrlZ             ('Z'-64)
#define CtrlF1            350
#define CtrlF2            351
#define CtrlF3            352
#define CtrlF4            353
#define CtrlF5            354
#define CtrlF6            355
#define CtrlF7            356
#define CtrlF8            357
#define CtrlF9            358
#define CtrlF10           359
#define CtrlAt            259
#define CtrlBackslash      28
#define CtrlBackspace     127
#define CtrlCaret          30
#define CtrlEnd           373
#define CtrlEnter          10
#define CtrlHome          375
#define CtrlLeft          371
#define CtrlLeftBracket    27
#define CtrlPgDn          374
#define CtrlPgUp          388
#define CtrlRight         372
#define CtrlRightBracket   29
#define CtrlUnderscore     31
#define ShiftF1           340
#define ShiftF2           341
#define ShiftF3           342
#define ShiftF4           343
#define ShiftF5           344
#define ShiftF6           345
#define ShiftF7           346
#define ShiftF8           347
#define ShiftF9           348
#define ShiftF10          349
#define ShiftTab          271
#define F1                315
#define F2                316
#define F3                317
#define F4                318
#define F5                319
#define F6                320
#define F7                321
#define F7                321
#define F8                322
#define F9                323
#define F10               324
#define Backspace           8
#define Del               339
#define Down              336
#define End               335
#define Enter              13
#define Esc                27
#define Home              327
#define Ins               338
#define Left              331
#define PgDn              337
#define PgUp              329
#define Right             333
#define Space              32
#define Tab                 9
#define Up                328

/* shift key states */

#define AltDown         0x08
#define CapsLockOn      0x40
#define CtrlDown        0x04
#define InsertOn        0x80
#define LeftShiftDown   0x02
#define NumLockOn       0x20
#define RightShiftDown  0x01
#define ScrollLockOn    0x10

#define absolute_disk_read(d,s,c,b)       absolute_disk_access(TRUE,d,s,c,b)
#define absolute_disk_read_far(d,s,c,b)   absolute_disk_access_far(TRUE,d,s,c,b)
#define absolute_disk_write(d,s,c,b)      absolute_disk_access(FALSE,d,s,c,b)
#define absolute_disk_write_far(d,s,c,b)  absolute_disk_access_far(FALSE,d,s,c,b)
#define command_line                      ((char *) 0x81)
#define command_line_length               (*(unsigned char *)0x80)
#define copyright_notice                  ((char *) 0x104)
#define feof(f)                           ((f)->flags&1)
#define ferror(f)                         ((f)->flags&2)
#define is_device(h)                      (device_information(h)&IS_DEVICE)
#define make_directory(s)                 mrs_directory(0x39,s)
#define offset(p)                         ((unsigned *)(&(p)))[0]
#define ready_to_send_byte(s)             (__inportb__((s)->base+5)&0x20)
#define remove_directory(s)               mrs_directory(0x3A,s)
#define segment(p)                        ((unsigned *)(&(p)))[1]
#define select_directory(s)               mrs_directory(0x3B,s)
#define send_byte(c,s)                    __outportb__((s)->base, c)

extern unsigned       absolute_disk_access(boolean, unsigned, unsigned, unsigned, void *);
extern unsigned       absolute_disk_access_far(boolean, unsigned, unsigned, unsigned, void far *);
extern void far      *allocate(unsigned);
extern void far      *allocate_permanent(unsigned);
extern unsigned       attribute_and_character_at_cursor(void);
extern unsigned       bdos(unsigned,...);
extern int            block_compare(void *, void *, unsigned);
extern int            block_compare_far(void far *, void far *, unsigned);
extern void           block_copy(void *, void *, unsigned);
extern void           block_copy_far(void far *, void far *, unsigned);
extern void           block_fill(int, void *, unsigned);
extern void           block_fill_far(int, void far *, unsigned);
extern filespecs_type clean_filespecs(char *);
extern void           clear_fast_ticks(void);
extern unsigned long  clock_ticks;
extern boolean        close(int);
extern boolean        close_putchar(void);
extern void           close_serial(struct serial *);
extern char          *command_line_argument(unsigned);
extern char           comparison_table[96];
extern void           context_switch(unsigned *);
extern unsigned       cpu_type(void);
extern int            create(char *, unsigned);
extern boolean        critical_error_flag;
extern char          *current_directory(unsigned);
extern unsigned       current_drive(void);
extern unsigned       cursor(void);
extern void           date_and_time(struct date_and_time *);
extern void           deallocate(void far *);
extern boolean        delete(char *);
extern unsigned       device_information(int);
extern void           disable_Ctrl_C(void);
extern void           display_at_cursor(unsigned);
extern unsigned       display_mode(void);
extern unsigned char  display_page;
extern unsigned       divide_and_remainder(unsigned long *, unsigned);
extern boolean        dos_command(char *);
extern unsigned       dos_version(void);
extern char          *edit_dollar(unsigned long, boolean);
extern char          *edit_hex(unsigned);
extern char          *edit_unsigned(unsigned, boolean);
extern char          *edit_unsigned_long(unsigned long, boolean);
extern void           enable_interrupts();
extern void           end_clock_ticks(void);
extern char far      *environment_string(char *);
extern unsigned       equipment(void);
extern unsigned       extended_key(void);
extern boolean        extended_keyboard_present(void);
extern void           extended_tsr_and_pop(unsigned);
extern unsigned       error_code(void);
extern void           error_exit(char *);
extern void           error_output(int);
extern void           error_output_string(char *);
extern boolean        execute_program(char *, char *, unsigned);
extern void           exit(unsigned);
extern void far      *farmalloc(unsigned long);
extern unsigned long  fast_ticks(void);
extern boolean        fclose(FILE *);
extern int            file_attributes(char *);
extern boolean        find_file(struct file_information *, char *, unsigned);
extern FILE          *fopen(char *, boolean, unsigned, jump_buffer, jump_buffer);
extern FILE          *fopen_handle(int, boolean, unsigned, jump_buffer, jump_buffer);
extern void           free(void *);
extern void           freeze_heap(void);
extern int            getc(FILE *);
extern int            getchar(void);
extern drive_type     get_drive_type(unsigned);
extern unsigned       heap_bottom;
extern unsigned       heap_top;
extern void far      *interrupt_vector(int);
extern void           insufficient_memory(void);
extern void           insufficient_table_memory(void);
extern void           jump(jump_buffer);
extern boolean        jumped(jump_buffer);
extern unsigned       key(void);
extern unsigned       key_states(void);
extern void           line_input(char *, char *, char *, unsigned);
extern void          *linked_list_sort(void *, unsigned, int (*)());
extern void far      *linked_list_sort_far(void far *, unsigned, int (*)());
extern unsigned       long file_date_and_time(int);
extern unsigned       long free_disk_space(unsigned);
extern unsigned       long length(int);
extern void          *malloc(unsigned);
extern unsigned       maximum_block_size;
extern unsigned       maximum_permanent;
extern unsigned       maximum_permanent;
extern boolean        mrs_directory(unsigned, char *);
extern boolean        next_file(struct file_information *);
extern int            open(char *, unsigned);
extern unsigned      *open_coroutine(void (f)(void), unsigned);
extern struct serial *open_serial(unsigned, unsigned, unsigned, unsigned, unsigned, unsigned, unsigned);
extern void           pop_up(unsigned, unsigned, unsigned, unsigned, unsigned *);
extern void           pop_up_slow(unsigned, unsigned, unsigned, unsigned, unsigned *);
extern boolean        print(int, unsigned);
extern boolean        print_serial(int, unsigned);
extern unsigned       printf(char *, ...);
extern char far      *program_filespecs(void);
extern char           program_name[];
extern void           putc(int, FILE *);
extern void           putchar(int);
extern boolean        putchar_error;
extern void           putchar_string(char *);
extern void           putc_string(char *, FILE *);
extern int            read(int, void *, int);
extern int            read_far(int, void far *, int);
extern boolean        reallocate(void far *, unsigned);
extern int            receive_byte(struct serial *s);
extern unsigned       redirected_key(void);
extern unsigned       redirected_key_eof(void);
extern unsigned       remaining_time;
extern boolean        rename(char *, char *);
extern void           scroll(unsigned, unsigned, int, unsigned);
extern boolean        seek(int, unsigned long, unsigned);
extern boolean        send(int, unsigned);
extern boolean        set_comparison_table(char *);
extern void           set_cursor(unsigned);
extern void           set_device_information(int, unsigned);
extern void           set_display_mode(unsigned);
extern boolean        set_file_attributes(unsigned, char *);
extern boolean        set_file_date_and_time(int, struct time, struct date);
extern void           set_interrupt_vector(unsigned, void far *);
extern void           set_timeout(unsigned);
extern void           set_up_critical_error(void);
extern void           sound_off(void);
extern void           sound_on(unsigned);
extern void           standard_output(int);
extern void           standard_output_string(char *);
extern void           start_clock_ticks(void);
extern void           start_fast_clock(void);
extern void           stop_fast_clock(void);
extern int            string_compare(char *, char *);
extern int            string_compare_far(char far *, char far *);
extern void           string_concatenate(char *, char *);
extern void           string_concatenate_far(char far *, char far *);
extern void           string_copy(char *, char *);
extern void           string_copy_far(char far *, char far *);
extern unsigned       string_length(char *);
extern unsigned       string_length_far(char far *);
extern int            tabular_string_compare(char *, char *);
extern int            tabular_string_compare_far(char far *, char far *);
extern boolean        timeout(void);
extern void           tsr_and_pop(unsigned);
extern void           tty(int);
extern int            write(int, void *, int);
extern int            write_far(int, void far *, int);
extern unsigned       __brklvl;

/* internal to the TT Library */

extern boolean        _no_error(void);

